/**
 * @module libflitter/app/FlitterApp
 */

const { Container, DependencyInjector } = require('flitter-di')
const Service = require('../NamedService')
const output = new (require('../utility/services/Output.service'))()
const Unit = require('../Unit')
const Analyzer = require('./Analyzer')

/**
 * The Flitter application.
 * @extends module:flitter-di/src/Service~Service
 */
class FlitterApp extends Service {
    /**
     * Get the name of this app's service: 'app'
     * @returns {string} - 'app'
     */
    static get name() {
        return 'app'
    }

    /**
     * If true, the {@link module:libflitter/app/Analyzer~Analyzer} will check
     * the application dependencies before run.
     * @type {boolean}
     */
    enable_analyzer = true

    /**
     * Instantiate the application.
     * @param {object} units - mapping of names to static Unit CLASS definitions
     */
    constructor(units) {
        super()

        /**
         * The app's units.
         * @type {Array<module:libflitter/Unit~Unit>}
         */
        this.__units = Array.isArray(units) ? units : Object.values(units)

        /**
         * Instance of the output service.
         * @type {module:libflitter/utility/services/Output~Output}
         */
        this.output = output

        /**
         * Array of unit names.
         * @type {Array<string>}
         * @private
         */
        this.__unit_array = []

        /**
         * Mapping of directory names to paths.
         * @type {object}
         */
        this.directories = {}

        this.__init_dependency_injector()

        /**
         * The underlying Express.js app.
         * @type {express}
         */
        this.express = require('express')()

        /**
         * Analyzer used to process various aspects of the application.
         * @type {module:libflitter/app/Analyzer~Analyzer}
         */
        this.analyzer = new Analyzer(this, units)
    }

    /**
     * Inject dependencies into the static class.
     * @param {module:flitter-di/src/Injectable~Injectable} Class - the static CLASS
     * @returns {*} - the injected CLASS
     */
    make(Class) {
        return this.__di.make(Class)
    }

    /**
     * Get the DI.
     * @returns {module:flitter-di/src/DependencyInjector~DependencyInjector}
     */
    di() {
        return this.__di
    }

    /**
     * Run the application by starting, then cleanly stopping all units.
     * @returns {Promise<void>}
     */
    async run() {
        await this.up().catch(e => { throw e })
        await this.down()
    }

    /**
     * Initialize the application by starting all units.
     * @returns {Promise<void>}
     */
    async up() {
        this.output.message('Starting Flitter...', 0)

        this.output.info('Checking Unit dependency order...')
        if ( this.enable_analyzer ) this.analyzer.check_dependencies()

        // Start the units
        for ( const unit_name of this.__unit_array ) {
            const unit = this.__di.service(unit_name)
            await this.__init_unit(unit)
        }
    }

    /**
     * Stop the application cleanly by stopping all units.
     * @returns {Promise<void>}
     */
    async down() {
        this.output.message('Stopping Flitter...', 0)

        for ( const unit_name of this.__unit_array ) {
            const unit = this.__di.service(unit_name)

            if ( unit.status() === Unit.STATUS_RUNNING ) {
                await this.__stop_unit(unit)
            }
        }
    }

    /**
     * Start a single unit.
     * @param {module:libflitter/Unit~Unit} unit
     * @returns {Promise<*>}
     * @private
     */
    async __init_unit(unit) {
        this.output.info(`Starting ${unit.name()}...`)

        try {
            unit.status(Unit.STATUS_STARTING)

            // register the unit's directories
            const directories = typeof unit.directories === 'function' ? unit.directories() : unit.directories
            this.directories = {...this.directories, ...directories}

            const go = await unit.go(this)
            unit.status(Unit.STATUS_RUNNING)
            return go
        } catch (e) {
            this.output.error(`Error encountered while starting ${unit.name()}!`)
            unit.status(Unit.STATUS_ERROR)
            throw e
            // TODO display error
        }
    }

    /**
     * Stop a single unit.
     * @param {module:libflitter/Unit~Unit} unit
     * @returns {Promise<void>}
     * @private
     */
    async __stop_unit(unit) {
        this.output.info(`Stopping ${unit.name()}...`)

        try {
            unit.status(Unit.STATUS_STOPPING)
            const cleanup = await unit.cleanup(this)
            unit.status(Unit.STATUS_STOPPED)
            return cleanup
        } catch (e) {
            this.output.error(`Error encountered while stopping ${unit.name()}.`)
            unit.status(Unit.STATUS_ERROR)
            throw e
        }
    }

    /**
     * Initialize the dependency injector for this app, adding the configured
     * units to the container as services.
     * @private
     */
    __init_dependency_injector() {
        const service_definitions = {
            app: this.constructor
        }

        for ( const unit of this.__units ) {
            service_definitions[unit.name] = unit
            this.__unit_array.push(unit.name)
        }

        this.__service_container = new Container(service_definitions)
        this.__service_container.register_as_instance('app', this)
        this.__di = new DependencyInjector(this.__service_container)
    }
}

module.exports = exports = FlitterApp
